/* ==================================================== */
/* @file: bali_mat.c					*/
/* scoring matrix io					*/
/* @history:						*/
/* @+	Apr. 97 <AV> first draft			*/
/* ==================================================== */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "Gtypes.h"
#include "opscan.h"
#include "bali.h"
#include "util.h"		/* for IS_UPPER	*/

#ifndef BUFSIZ
#define BUFSIZ       1024
#endif

/* -------------------------------------------- */
/* get char position in alpha	     		*/
/* -------------------------------------------- */
static int sGetCharPos(char *alpha, int c)

{
    char *p;
    
    if (! (p = strchr(alpha, c)))
	return (c == '*' ? -1 : sGetCharPos(alpha, '*'));
    
    return (int) (p - alpha);
}

/* -------------------------------------------- */
/* normalize matrix & compute weights		*/
/* -------------------------------------------- */

static void sNormalizeMatrix(ScorMat *mat)
{
    int	    i, j, ipos, jpos, code, val;
    float   fval, fnorm;
 
    fnorm = (float) (mat->smax - mat->smin);

    for (i = 0 ; i < mat->nalphabet ; i++) {
	ipos = sGetCharPos(mat->alphabrut, 'A' + i);
			/* bypass non common character	*/
	for (j = 0 ; j < mat->nalphabet ; j++) {
			/* bypass non common character	*/
	    jpos = sGetCharPos(mat->alphabrut, 'A' + j);   
	    code = HCODE_INDEX(i, j);
	    val  = (i == j ? mat->smax : mat->brutscore[ipos][jpos]);
	    val  = MIN2(val, mat->smax);
	    val  = MAX2(val, mat->smin);


	    fval =  (float) (val - mat->smin) / fnorm ;

				/* <jompo> similarity score	*/
				/* between 0.0 and 1.0		*/
	    mat->simscore[code] = fval;

				/* distance score	*/
	    mat->weight[code] = SCORENORM - (int) floor((double) fval * SCORENORM + 0.5);


	}
    }

    return;
} 
 
/* -------------------------------------------- */
/* read matrix (blast format)			*/
/* -------------------------------------------- */

int ReadScoreMatrix(char *filename, ScorMat *mat)
{
    int	    i, j, ipos, jpos, code, val;
    char    *pb, buffer[BUFSIZ];
    
    if (! AssignToStdin(filename))
	return 0;
	
    mat->nalphabrut = 0;

    mat->nalphabet = NSTDALPHABET;
    (void)strcpy(mat->alphabet, STDALPHABET);

    /* ---------------- */
    /* read alphabet	*/

    while (gets(buffer)) {

	if (*buffer == '#')
	    continue;

	/* read alphabet */	    

	for (pb = buffer ; *pb ; pb++) {
	    if (IS_UPPER(*pb) || (*pb == '*'))
		mat->alphabrut[mat->nalphabrut++] = *pb;  
	}

	break;
    }
    
    mat->alphabrut[mat->nalphabrut] = '\000';

    /* ---------------- */
    /* check alphabet	*/
    
    for (i = 0 ; i < mat->nalphabrut ; i++) {
	if (sGetCharPos(mat->alphabrut, 'A' + i) < 0)
	    return Erreur("[ReadScoreMatrix] invalid alphabet", 0);
    }
    
    /* ---------------- */
    /* read matrix	*/
    
    for (i = 0 ; i < mat->nalphabrut ; i++) {
	for (j = 0 ; j < mat->nalphabrut ; j++) {
	    if (scanf("%d", &val) != 1)
		return Erreur("[ReadScoreMatrix] matrix io error", 0);
	    mat->brutscore[i][j] = val;	    
	}
    }

#if 0
    /* verif	*/
    fprintf(stderr,"Matrix alpha=%s\n",mat->alphabrut);
    for (i = 0 ; i < mat->nalphabrut ; i++) {
       for (j = 0 ; j < mat->nalphabrut ; j++) {
	  fprintf(stderr,"%3d",mat->brutscore[i][j]);
       }
       fprintf(stderr,"\n");
    }
#endif

    /* ---------------- */
    /* get min and max	*/
    
    mat->smin = mat->smax = mat->brutscore[0][1];
    for (i = 0 ; i < mat->nalphabrut ; i++) {
	for (j = i ; j < mat->nalphabrut ; j++) {
	    mat->smin = MIN2(mat->smin,  mat->brutscore[i][j]);
	    mat->smax = MAX2(mat->smax,  mat->brutscore[i][j]);
	}
    }

    /* ---------------- */
    /* set score indexed on ascii alphabet	*/

    for (i = 0 ; i < mat->nalphabet ; i++) {
        ipos = sGetCharPos(mat->alphabrut, 'A' + i);
	if (! strchr(mat->alphabet, mat->alphabrut[i]))
	  continue;
	for (j = 0 ; j < mat->nalphabet ; j++) {
	   if (! strchr(mat->alphabet, mat->alphabrut[j]))
	      continue;
	   jpos = sGetCharPos(mat->alphabrut, 'A' + j);
           code = HCODE_INDEX(i, j);
	   mat->alphascore[code] =  mat->brutscore[ipos][jpos];
        }
    }

    /* ---------------- */
    /* normalize	*/

    sNormalizeMatrix(mat);

    return 1;
}

/* -------------------------------------------- */
/* identity matrix (blast format)		*/
/* -------------------------------------------- */

int IdentityMatrix(ScorMat *mat)
{
    int i, j;

    mat->nalphabrut = mat->nalphabet = NSTDALPHABET;

    for (i = 0 ; i < mat->nalphabet ; i++) {
	mat->alphabet[i] = mat->alphabrut[i] = 'A' + i;
	for (j = 0 ; j < mat->nalphabet ; j++) {
	   mat->alphascore[HCODE_INDEX(i, j)] = 
	     mat->brutscore[i][j] = (i == j ? 1 : -1);
	}
    }
    mat->alphabet[NSTDALPHABET] = '\000';	

    mat->smax = 1;
    mat->smin = -1;
    
    sNormalizeMatrix(mat);
   
    return 1;
}


/* -------------------------------------------- */
/* Write matrix to fout				*/
/* -------------------------------------------- */
void PrintMatrix(ScorMat *mat, FILE *fout)
{
   int i, j;

   if (! mat) {
     fprintf(fout,"Null Matrix  !!!\n");
     return;
   }
   
   fprintf(fout,"INITIAL SCORES \n");
   fprintf(fout,"  ");
   for (i = 0 ; i < mat->nalphabrut ; i++)
     fprintf(fout,"  %c ", mat->alphabrut[i]);
   fprintf(fout,"\n");

   for (i = 0 ; i < mat->nalphabrut ; i++) {
	fprintf(fout,"%c ", mat->alphabrut[i]);
	for (j = 0 ; j < mat->nalphabrut ; j++)
	    fprintf(fout,"%3d ", mat->brutscore[i][j]);
        fprintf(fout,"\n");
   }

   fprintf(fout,"INITIAL SCORES (ASCII ranked)\n");
   fprintf(fout,"  ");
   for (i = 0 ; i < mat->nalphabet ; i++) {
     if (! strchr(STDALPHABET, mat->alphabet[i]))
       continue;
     fprintf(fout,"  %c ", (char)('A' + i));
   }
   fprintf(fout,"\n");

   for (i = 0 ; i < mat->nalphabet ; i++) {
        if (! strchr(STDALPHABET, mat->alphabet[i]))
          continue;
	fprintf(fout,"%c ", mat->alphabet[i]);
	for (j = 0 ; j < mat->nalphabet ; j++) {
            if (! strchr(STDALPHABET, mat->alphabet[i]))
               continue;
	    fprintf(fout,"%3d ", mat->alphascore[HCODE_INDEX(i,j)]);
	}
        fprintf(fout,"\n");
   }

   fprintf(fout,"SIMILARITY NORMALIZED SCORES \n");
   fprintf(fout,"  ");
   for (i = 0 ; i < mat->nalphabet ; i++) {
     if (! strchr(STDALPHABET, mat->alphabet[i]))
       continue;
     fprintf(fout,"  %c  ", mat->alphabet[i]);
   }
   fprintf(fout,"\n");

   for (i = 0 ; i < mat->nalphabet ; i++) {
        if (! strchr(STDALPHABET, mat->alphabet[i]))
          continue;
	fprintf(fout,"%c ", mat->alphabet[i]);
	for (j = 0 ; j < mat->nalphabet ; j++) {
            if (! strchr(STDALPHABET, mat->alphabet[i]))
               continue;
	    fprintf(fout,"%4.2f ", mat->simscore[HCODE_INDEX(i,j)]);
	}
        fprintf(fout,"\n");
   }

   return;

}
