/* ==================================================== */
/* Gloup Aug. 95					*/
/* @file: fastp_core.c					*/
/* fastp core						*/
/* history:						*/
/* Aug. 95 <gloup> first draft				*/
/* Jul 99  <jompo> another draft			*/
/* ==================================================== */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Gtypes.h"
#include "util.h"
#include "fastp_hash.h"
#include "fastp_core.h"


/* ------------------------------------------------ */
/* liberation du tableau des diags		    */
/* ------------------------------------------------ */
DiagTable *FreeDiagTable(DiagTable *diags)
{
    if (diags) {
	if (diags->diag)
	    FREE(diags->diag);
	FREE(diags);
    }

    return NULL;
}

/* ------------------------------------------------ */
/* allocation du tableau des diags		    */
/* ------------------------------------------------ */
DiagTable *NewDiagTable(int diag_size)
{
    DiagTable *diags; 

    if (! (diags = NEW(DiagTable))) {
	(void) MemoryErreur("[NewDiagTable]", ErrMemory);
	return NULL;
    }
    
    diags->diag_size   = diag_size;
    diags->diag_offset = 0;
    diags->diag_integ  = 1;
    
    if (! (diags->diag = NEWN(int, diag_size))) {
	(void) MemoryErreur("[NewDiagTable]", ErrMemory);
	return FreeDiagTable(diags);
    }

    (void) memset(diags->diag, 0, diag_size * sizeof(int));
    
    return diags; 
}

/* ------------------------------------------------ */
/* calcul et remplissage du tableau des diagonales  */
/* ------------------------------------------------ */
void ComputeDgnl(char *cseq, HashTable *hash, DiagTable *diags)
{
    int		pos, *pdiags;
    Int32	h;
    Cell	*curr;
    HashHeader	*header;
    char	*s;

    header = &(hash->header);	    
					/* remise a zero		*/

   pdiags = diags->diag;
   for (pos = 0 ; pos < diags->diag_size ; pos++)
	*pdiags++ = 0;

					/* ---------------------------- */
      					/* codage et remplissage des 	*/
      					/* diagonales			*/
					/* ---------------------------- */

    h = HashValue(cseq, header);
    
    for (pos = 0, s = cseq + header->ksize ; *s != EOS ; pos++, s++) {

	for (curr = hash->table[h] ; curr ; curr = curr->next)
	    diags->diag[pos - curr->pos + diags->diag_offset]++;

	h = UpdateHashValue(h, *s, header);
    }


}

/* -------------------------------------------------------------------- */
/* integration/selection des diagonales					*/
/* en cas de doublon on choisit la diagonale la plus proche de la	*/
/* diagonale principale							*/
/* -------------------------------------------------------------------- */
int IntegrateDgnl(DiagTable *diags, int* score)
{
    int	i, imax, somme, delta, 
	max_somme, max_diag, min_delta, middle, max_middle;
    int	*pdiag, *p_old_diag;

#if 0 /* debug jompo */
{
	int kk;
	fprintf(stderr, "jompo\n");
	for(kk=0;kk<diags->diag_size;kk++)
   		fprintf(stderr,"diag %d %d\n",kk, diags->diag[kk]);
}
#endif

    pdiag = p_old_diag = diags->diag;
    somme = *pdiag++;

    for (i = 1; i <= 2*diags->diag_integ ; i++)
	somme += *pdiag++;

    max_somme = somme;
    max_diag  = 0 + diags->diag_integ;
    max_middle = diags->diag[0+diags->diag_integ];
    min_delta = diags->diag_offset;
    imax      = diags->diag_size - diags->diag_integ;

#if 0 /* debug jompo */
    fprintf(stderr,"\nimax = %d\n",imax);
#endif

    for (i = diags->diag_integ + 1 ; i <= imax ; i++) {

	
	somme += *pdiag++ - *p_old_diag++;

	if (somme < max_somme)
	    continue;


	delta = ABS(i-diags->diag_offset);
	middle = diags->diag[i];

			/* si ex-aequo lointains ou proches	*/
			/* on garde :				*/
			/* 1) celle qui a le plus haut "milieu"	*/
			/* 2) la plus proche de la diagonale	*/	
	if (somme == max_somme) {
			/* si pic du milieu < pic milieu du	*/
			/* precedent ex-aequo, continue		*/
	   if (middle < max_middle)
	      continue;
	   else if (middle == max_middle && delta > min_delta)
	      continue;
	}

	max_somme = somme;
	max_diag  = i;
	max_middle = middle;
	min_delta = delta;
    }
    if (score)
	*score = max_somme;


#if 0 /* debug jompo */
{
	fprintf(stderr, "\nmax_somme = %d\n", max_somme);
	fprintf(stderr, "*score = %d\n", *score);
	fprintf(stderr, "max_diag = %d\n", max_diag);
	fprintf(stderr, "max_middle = %d\n", max_middle);
}
#endif

    return max_diag;
}

	
/* -------------------------------------------------------------------- */
/* compute alignment score in shifting s1 from offset 			*/
/* in respect to s2, and summing the individual scores			*/
/* acording to ScorMat pmatrix						*/
/* NOTE: sequence are indexed						*/
/* NOTE2: score is floating point value between 0.0 and 1.0		*/
/*   and is computed as sum of individual score along alignment divided	*/
/*   by the length of the smaller sequence				*/
/* -------------------------------------------------------------------- */
float Aliscore(char *s1, char *s2, int len1, int len2, int offset, ScorMat *pmatrix)
{
	float	score;
	int	i, len_ali, min_len;
#undef DEBALI
#ifdef DEBALI
	char *ss1,*ss2, *ss1p, *ss2p;
	int	lim;
	char	c1,c2;
#endif

	score = 0.0;
	min_len = MIN(len1,len2);

	if (offset < 0) {
	   s1   -= offset;
	   len1 += offset;
	}
	else {
	   s2   +=  offset;
	   len2 -= offset;
	}

	len_ali = MIN(len1,len2);

#ifdef DEBALI
	ss1p = ss1 = NEWN(char, len_ali+1);
	ss2p = ss2 = NEWN(char, len_ali+1);
#endif

	for (i = 0; i < len_ali; i++) {
#ifdef DEBALI
	   *ss1p++ = (char)(*s1) + 'A';
	   *ss2p++ = (char)(*s2) + 'A';
#endif
	   score += pmatrix->simscore[HCODE_INDEX(*s1++,*s2++)];
	   /* score += pmatrix->simscore[*s1++ << 5 | *s2++]; */
	}

#ifdef DEBALI
#define LENLINE 60
	*ss2p = *ss1p = '\0';
	ss2p = ss2;
	ss1p = ss1;
	lim = (int)strlen(ss1)/(int)LENLINE + 1;

	fprintf(stderr, "\n\n");
        for (i = 0; i < lim; i++) {
	   ss1p = ss1 + LENLINE*i;
	   ss2p = ss2 + LENLINE*i;
	   if ((int)strlen(ss1) >= LENLINE) {
	      c1 = *(ss1p+LENLINE);
	      c2 = *(ss2p+LENLINE);
	      *(ss1p+LENLINE) = '\0';
	      *(ss2p+LENLINE) = '\0';
	   }
	   else
	      c1 = c2 = 'Z';

	   fprintf(stderr, "1  : %s\n", ss1p);
	   fprintf(stderr, "2  : %s\n", ss2p);
	   
	   if (c1 != 'Z' && c2 != 'Z') {
	      *(ss1p+LENLINE) = c1;
	      *(ss2p+LENLINE) = c2;
	   }
	   fprintf(stderr, "\n");

	}

	FREE(ss1);
	FREE(ss2);
#endif
	return score/(float)min_len;
}


/* -------------------------------------------------------------------- */
/* compute alignment score in finding maximum scoring segment		*/
/* (HSSP) in shifting s1 from offset 					*/
/* in respect to s2, and summing the individual scores			*/
/* acording to ScorMat pmatrix						*/
/* NOTE: input sequence are indexed					*/
/* NOTE2: score is floating point value between 0.0 and 1.0		*/
/*   and is computed as sum of individual score along the		*/
/*   best sub-sum scoring region of shifted sequence divided		*/
/*   by the length of the smaller sequence				*/
/* -------------------------------------------------------------------- */
float HspScore(char *s1, char *s2, int len1, int len2, int offset, ScorMat *pmatrix)
{
	int	i, len_ali, min_len;
        float	maxsofar, maxendinghere, pair_score, score;
#define DEBALI
#ifdef DEBALI
				/* start and stop of region     */
        int maxsofstart, maxsofstop, maxendstart, maxendstop;
	char *ss1,*ss2, *ss1p, *ss2p;
	int	lim;
	char	c1,c2;
#endif

	min_len = MIN(len1,len2);

	if (offset < 0) {
	   s1   -= offset;
	   len1 += offset;
	}
	else {
	   s2   +=  offset;
	   len2 -= offset;
	}

	len_ali = MIN(len1,len2);

#ifdef DEBALI
	ss1p = ss1 = NEWN(char, len_ali+1);
	ss2p = ss2 = NEWN(char, len_ali+1);
#endif

	maxsofstart   = maxsofstop = 0;
	maxsofar      = 0.0;
	maxendstart   = maxendstop = 0;
	maxendinghere = 0.0;

	for (i = 0; i < len_ali; i++,s1++,s2++) {
#ifdef DEBALI
	   *ss1p++ = (char)(*s1) + 'A';
	   *ss2p++ = (char)(*s2) + 'A';
#endif
	   pair_score = (float)(pmatrix->alphascore[HCODE_INDEX(*s1,*s2)]);
	   if (maxendinghere + pair_score > 0.0) {
		maxendinghere += pair_score;
		maxendstop = i;
	   }
	   else {
	      maxendinghere = 0.0;
	      maxendstart = maxendstop = i + 1;
	   }
	   maxsofar = (maxsofar > maxendinghere) ? maxsofar : maxendinghere;
#ifdef DEBALI
	   maxsofstart = (maxsofar > maxendinghere) ? maxsofstart : maxendstart;
	   maxsofstop  = (maxsofar > maxendinghere) ? maxsofstop  : maxendstop;
#ifdef DEBALO
	   fprintf(stderr,"maxsofar->%f maxendinghere->%f (score %c|%c val %f)\n",
		maxsofar, maxendinghere, *s1+'A', *s2+'A', pair_score);
#endif
#endif
	}

#ifdef DEBALI
#define LENLINE 60
	*ss2p = *ss1p = '\0';
	ss2p = ss2;
	ss1p = ss1;
	lim = (int)strlen(ss1)/(int)LENLINE + 1;
        for (i = 0; i < len_ali; i++,ss1p++,ss2p++) {
	   if (i < maxsofstart || i > maxsofstop) {
	      *ss1p = (char)tolower((int)*ss1p);
	      *ss2p = (char)tolower((int)*ss2p);
           }
	}
	ss2p = ss2;
	ss1p = ss1;

	fprintf(stderr, "\nmxscore %f\n", maxsofar/(float)min_len);
        for (i = 0; i < lim; i++) {
	   ss1p = ss1 + LENLINE*i;
	   ss2p = ss2 + LENLINE*i;
	   if ((int)strlen(ss1) >= LENLINE) {
	      c1 = *(ss1p+LENLINE);
	      c2 = *(ss2p+LENLINE);
	      *(ss1p+LENLINE) = '\0';
	      *(ss2p+LENLINE) = '\0';
	   }
	   else
	      c1 = c2 = 'Z';

	   fprintf(stderr, "1  : %s\n", ss1p);
	   fprintf(stderr, "2  : %s\n", ss2p);
	   
	   if (c1 != 'Z' && c2 != 'Z') {
	      *(ss1p+LENLINE) = c1;
	      *(ss2p+LENLINE) = c2;
	   }
	   fprintf(stderr, "\n");

	}

	FREE(ss1);
	FREE(ss2);
#endif

#undef MAX_MATCH_MATCH

		/* scores */
	/* return (maxsofar/(float)pmatrix->smax/(float)min_len); */
	/* return (maxsofar/(float)pmatrix->smax)/(maxsofstop-maxsofstart)*(len_ali/(float)min_len); */
	/* return (maxsofar/(float)pmatrix->smax)*(len_ali/(float)min_len); */
	/* return (maxsofar/(float)pmatrix->smax)/(float)len_ali; */
				/* normalize score by higher score in	*/
				/* matrix, and by protein mean length 	*/
				/* in order to have something between	*/
				/* 0.0 and 1.0				*/
				/* if > 1.0, set to 1.0			*/
	score = (maxsofar/(float)pmatrix->smax)/300.0;

		/* jompo a finir: mettre return (score > 1.0) ? 1.0 : score	*/
	return score;

}


