/* ==================================================== */
/* Gloup Aug. 95					*/
/* @file: fastp_hash.c					*/
/* sequence indexing and hashing			*/
/* history:						*/
/* Aug. 95 <gloup> first draft				*/
/* ==================================================== */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Gtypes.h"
#include "opscan_help.h"
#include "fastp_hash.h"
#include "util.h"

/* -------------------------------------------- */
/* @static					*/
/* symbol error					*/
/* -------------------------------------------- */
static char *sSymbolError(char *seq)
{
    char buffer[BUFSIZ];
    
    sprintf(buffer, "Bad symbol >%c< in sequence", *seq);
    
    (void) Erreur(buffer, 0);
    
    return NULL;
}

/* -------------------------------------------- */
/* sequence indexing (in seq buffer)	 	*/
/* -------------------------------------------- */

char *IndexSequence(char *seq, char *alpha)
{
    char *s, *f;

    for (s = seq ; *s ; s++) {
	if (! (f = strchr(alpha, *s)))
	    return sSymbolError(s);
	*s = (signed char) (f - alpha);
    }

    *s = EOS;
    
    return seq;
}

/* -------------------------------------------- */
/* sequence deindexing (from seq buffer)	*/
/* -------------------------------------------- */

char *DeIndexSequence(char *seq,  char *alpha)
{
    char *s = seq;

    while (*s != EOS) {
	*s = alpha[(int)(*s)];
	s++;
    }

    *s = '\000';

    return seq;
}

/* -------------------------------------------------------------------- */
/* @static								*/
/* fonction d'ajout cellule dans une liste chainee			*/
/* -------------------------------------------------------------------- */

static void sAddInList(Cell *ref, Cell *cell)
{
    Cell *curr = ref;
	
    while(curr->next) 
	curr = curr->next;
	     
    curr->next = cell;

}

/* -------------------------------------------------------------------- */
/* fonction d'ajout cellule dans une liste chainee avec creation de 	*/
/* la cellule								*/
/* -------------------------------------------------------------------- */

Cell *PutInList(Cell **root, int pos)
{
    Cell *new;

    if (! (new = NEW(Cell))) {
	(void) MemoryErreur("[PutInList]", ErrMemory);
	return NULL;
    }
    
    new->pos  = pos;
    new->next = NULL;
	
    if (*root)
	sAddInList(*root, new);
    else
	*root = new;

    return *root;
	
}

/* -------------------------------------------------------------------- */
/* fonction de destruction d'une liste chainee				*/
/* -------------------------------------------------------------------- */
Cell *FreeList(Cell *root)
{
    Cell *curr, *save;
	
    for (curr = root ; curr ; curr = save) {
	save = curr->next;	      
	FREE(curr);
    }

    return NULL;
}

/* ------------------------------------------------ */
/* compute hashheader				    */
/* ------------------------------------------------ */

int ComputeHashHeader(HashHeader *header, char *alphabet, int ksize)
{
    int		i, bitsize;
    Int32	lk, pk;

    strcpy(header->alphabet, alphabet);
    
    header->ksize = ksize;
    
    header->alphasize = strlen(alphabet);
    
		/* compute bitshift = log2(alphasize) */

    lk = 0;
    for (pk = 1 ; pk < header->alphasize ; pk *= 2) 
	lk++;

    header->bitshift = lk;
    header->wsize    = pk;
   
    bitsize = header->ksize * header->bitshift;
    
    if (bitsize >= sizeof(Int32) * 8) {
	(void) Erreur("k-uple too big with this alphabet", 20); 
	return 0;
    }

		/* compute masks				*/
    lk = 0x1;
    for (i = 1 ; i < bitsize ; i++)
	lk = (lk << 1) | 0x1;
	
    header->wmask = lk;

    lk = 0x1;
    for (i = 1 ; i < header->bitshift ; i++)
	lk = (lk << 1) | 0x1;
	
    header->smask = lk;
   
		/* compute table size  = wsize^ksize	*/
    
    pk = 1;
    for (i = 0 ; i < header->ksize ; i++)
	pk *= header->wsize;

    header->tsize = pk;
    
    return 1;
}

/* ------------------------------------------------ */
/* free hashtable				    */
/* ------------------------------------------------ */
HashTable *FreeHashTable(HashTable *hash)
{
    int i;
    
    if (hash) {
	if (hash->table) {
	    for (i = 0 ; i < hash->header.tsize ; i++)
		(void) FreeList(hash->table[i]);
	    FREE(hash->table);
	}
	FREE(hash);
    }
    
    return NULL;
}

/* ------------------------------------------------ */
/* built hashtable for k length words		    */
/* ------------------------------------------------ */
HashTable *NewHashTable(char *alphabet, int ksize)
{
    int		i;
    HashTable	*hash;

    if (! (hash = NEW(HashTable))) {
	(void) MemoryErreur("[NewHashTable]", ErrMemory);
	return NULL;
    }

    hash->table = NULL;
    
    if (! ComputeHashHeader(&(hash->header), alphabet, ksize))
	return FreeHashTable(hash);
	
		/* allocate memory			 */
		
    hash->table = NEWN(Cell*, hash->header.tsize);
    
    if (! hash->table) {
	(void) MemoryErreur("[NewHashTable]", ErrMemory); 
	return FreeHashTable(hash);
    }
    
    for (i = 0 ; i < hash->header.tsize ; i++) 
	hash->table[i] = NULL;
    
    return hash;
}

/* ------------------------------------------------ */
/* reset hashtable				    */
/* ------------------------------------------------ */
void ResetHashTable(HashTable *hash)
{
    int i;
    
    for (i = 0 ; i < hash->header.tsize ; i++) {
	(void) FreeList(hash->table[i]);
	hash->table[i] = NULL;
    }
}

/* ------------------------------------------------ */
/* free hashvector				    */
/* ------------------------------------------------ */
HashVector *FreeHashVector(HashVector *hash)
{
    
    if (hash) {
	if (hash->counts)
	    FREE(hash->counts);
	FREE(hash);
    }

    return NULL;
}

/* ------------------------------------------------ */
/* built hashvector for k length words		    */
/* ------------------------------------------------ */
HashVector *NewHashVector(char *alphabet, int ksize)
{
    int		i;
    HashVector	*hash;

    if (! (hash = NEW(HashVector))) {
	(void) MemoryErreur("[NewHashVector]", ErrMemory);
	return NULL;
    }

    hash->counts = NULL;
    
    if (! ComputeHashHeader(&(hash->header), alphabet, ksize))
	return FreeHashVector(hash);
	
		/* allocate memory			 */
		
    hash->counts = NEWN(Int32, hash->header.tsize);
    
    if (! hash->counts) {
	(void) MemoryErreur("[NewHashVector]", ErrMemory); 
	return FreeHashVector(hash);
    }
    
    for (i = 0 ; i < hash->header.tsize ; i++) 
	hash->counts[i] = 0;
    
    return hash;
}

/* ------------------------------------------------ */
/* reset hashvector				    */
/* ------------------------------------------------ */
void ResetHashVector(HashVector *hash)
{
    int i;
    
    for (i = 0 ; i < hash->header.tsize ; i++)
	hash->counts[i] = 0;
}


/* ------------------------------------------------ */
/* compute hash value				    */
/* ------------------------------------------------ */
Int32 HashValue(char *word, HashHeader *header)
{
    int i, h;
    
    h = word[0];
    
    for (i = 1 ; i < header->ksize ; i++) {
	h <<= header->bitshift;  
	h |= (int) word[i];
    }
     
    return h;
}

/* ------------------------------------------------ */
/* update hash value				    */
/* should be inline				    */
/* ------------------------------------------------ */
Int32 UpdateHashValue(Int32 old, int newsymb, HashHeader *header)
{
    return (((old << header->bitshift) & header->wmask) | newsymb);
}


/* ------------------------------------------------ */
/* fonction de hashing simple			    */
/* ------------------------------------------------ */
int HashUserSeq(HashTable *hash, char *cseq)
{
    int		pos;
    Int32	h;
    HashHeader	*header;
    char	*s;
    
    header = &(hash->header);

    h = HashValue(cseq, header);
    pos = 0;
        
    for (s = cseq + header->ksize ; *s != EOS ; s++) {
   
	PutInList(hash->table + h, pos++);

	h = UpdateHashValue(h, *s, header);
    }
    
    return pos; 
}

/* ------------------------------------------------ */
/* fonction de comptage simple			    */
/* ------------------------------------------------ */
int CountUserSeq(HashVector *hash, char *cseq)
{
    int		total;
    Int32	h;
    HashHeader	*header;
    char	*s;
    
    header = &(hash->header);

    h = HashValue(cseq, header);
    total = 0;
        
    for (s = cseq + header->ksize ; *s != EOS ; s++) {
	hash->counts[h]++;   
	h = UpdateHashValue(h, *s, header);
	total++;
    }  

    return total; 
}

/* ------------------------------------------------ */
/* score upper bound				    */
/* ------------------------------------------------ */
int UpperBoundScore(HashVector *hashi, HashVector *hashj)
{
    int	i,  total;

    for (i = total = 0 ; i < hashi->header.tsize ; i++)
	total += MIN(hashi->counts[i], hashj->counts[i]);

    return total;
}

