/* ------------------------------------------------------------	*/
/* operon.c functions on operon structures			*/
/* ------------------------------------------------------------	*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"
#include "operon.h"
#include "tags.h"

#define ABS(x)              ((x) >= 0  ? (x) : -(x))
#define MIN(a, b)           ((a) < (b) ? (a) : (b))

/* -------------------------------------------- */
/* fonctions statiques				*/
/* -------------------------------------------- */

static void sSetOneLastHomolog(OperonSeqPtr ope, float len_thresh,
		int seuil_score, int bijectif);

/* -------------------------------------------- */
/* OperonParseInfo : parse the information in	*/
/* the comment of the FastaSequence		*/
/* if standard_fasta_format == 1, the sequence	*/
/* is in standard fasta format, else it is	*/
/* in "mnemo format" (see README.opscan)	*/
/* --------------------------------------------	*/
static Bool OperonParseInfo(OperonSeqPtr ope, int standard_fasta_format) 
{
   int dummy;
   static int	counter = 0, deja_erreur = 0;


   if (ope == NULL)
       return Faux;

   if ((int)(strlen(ope->fseqptr->name)) >= MNEMOCHAR) {

        MyError(stderr, 0, REMARK_TAG, "OperonParseInfo:: error name %s is longer than MNEMOCHAR (%d)\n",
		ope->fseqptr->name, MNEMOCHAR);

        strncpy(ope->mnemo, ope->fseqptr->name, MNEMOCHAR-1);
	ope->mnemo[MNEMOCHAR-1] = '\0';
   }
   else {

        (void)strcpy(ope->mnemo, ope->fseqptr->name);

   }

   counter++;

   if (standard_fasta_format) {
        ope->ignore = counter;
	ope->strand = UNK_STRAND;
   }
   else if (sscanf(ope->fseqptr->comment, "%d %d %d %c", &(ope->ignore), &dummy, &dummy, &(ope->strand)) != 4) {

        if (deja_erreur == 0) {
	   MyError(stderr, 0, REMARK_TAG,
	     "OperonParseInfo:: error parsing %s\nshould be int, int, int, char\n",
		ope->fseqptr->comment);
	   MyError(stderr, 0, REMARK_TAG,
	     " Continue reading assuming standard fasta format...\n");
           deja_erreur = 1;
	}

	if (deja_erreur == 1)
	    ope->ignore = counter;
	else
	   ope->ignore = UNK_IGNORE;

	ope->strand = UNK_STRAND;
   }


				/* ignore is 0 to ignore operon	*/
				/* else -1 for unknown ignore	*/
				/* else positive if operon is	*/
				/* not to be ignored		*/
				/* it can be the rel_number of	*/
				/* the operon gene		*/
   if (ope->ignore == 0)
      ope->ignore = IGNORE;

   if (ope->strand != 'C' && ope->strand != 'D')
      ope->strand = UNK_STRAND;

   return Vrai;

}


/* -------------------------------------------- */
/* allocation d'un operon                       */
/* -------------------------------------------- */

OperonSeqPtr NewOperonSequence(int gen_kept)
{

   OperonSeqPtr ope;
   int		i;

   if (! (ope = NEW(OperonSequence)))
      return NULL;

   if (! (ope->fseqptr = NewFastaSequence())) {
      MyError(stderr,0,
	ERROR_TAG, "NewOperonSequence:: cannot create NewFastaSequence\n");
      free(ope);
      return NULL;
   }

   if (! (ope->gen_kept = NEWN(GenKept, gen_kept)) ) {
      MyError(stderr,0,
	ERROR_TAG, "NewOperonSequence:: cannot create gen_kept\n");
      free(ope->fseqptr);
      free(ope);
      return NULL;
   }

   ope->kept_by		= 0;
   ope->max_gen_kept	= gen_kept;
   ope->actual_gen_kept	= 0;
   ope->class	        = NO_CLASS;

   for (i = 0; i < ope->max_gen_kept; i++) {
      ope->gen_kept[i].gen_ptr = NULL;
      ope->gen_kept[i].score = (int)0;
      ope->gen_kept[i].stat = -1.0;
   }

   ope->homolog.score   = (int)0;
   ope->homolog.stat   = -1.0;
   ope->homolog.gen_ptr = NULL;

   ope->ignore = UNK_IGNORE;
   ope->strand = UNK_STRAND;
   ope->mnemo[0] = '\000';

   return ope;
}


/* -------------------------------------------- */
/* liberation d'un OperonSequence              */
/* -------------------------------------------- */

void FreeOperonSequence(OperonSeqPtr ope)
{
   if (ope) {

      if (ope->gen_kept)
	 free(ope->gen_kept);

      if (ope->fseqptr)
	 FreeFastaSequence(ope->fseqptr);

      free(ope);
   }

   return;
}
 
/* -------------------------------------------- */
/* lecture/redimensionnement d'un operon        */
/* <Lecture Serie>                              */
/* -------------------------------------------- */

Bool ReadOperonSequence(FILE *streamin, OperonSeqPtr ope, int standard_fasta_format) 
{

   if (ope == NULL) {
       MyError(stderr, 0,
	  ERROR_TAG, "ReadOperonSequence:ope is NULL");
       return Faux;
   } 

   if (ReadFastaSequence(streamin, ope->fseqptr) == Faux) {
       return Faux;
   }

   if (OperonParseInfo(ope, standard_fasta_format) == Faux) {
     MyError(stderr, 0,
       ERROR_TAG, "ReadOperonSequence:cannot parse %s", ope->fseqptr->comment);
     return Faux;
   }

   return Vrai;
}



/* -------------------------------------------- */
/* WriteOperonSequence: ecrit la sequence	*/
/* de genome sur stream				*/
/* -------------------------------------------- */
void WriteOperonSequence(FILE *stream, OperonSeqPtr ope, int len_line)
{

   (void)fprintf(stream, ">%s %d %d %d %c\n", ope->mnemo, ope->ignore, 0, 0, ope->strand);

   WriteFastaSequence(stream, ope->fseqptr, len_line, ONLY_SEQ);

   return;

}

/* -------------------------------------------- */
/* KeepGeneInOpeF: garde le pointeur sur le gene	*/
/* p_gen, de score score dans la tableau	*/
/* gen_kept de l'operon				*/
/* -------------------------------------------- */
void  KeepGeneInOpeF(OperonSeqPtr p_ope, GenomeSeqPtr p_gen, float fscore)
{
 int i, j, fini;

			/* -----------------------------------	*/
			/* if fscore is inf. to min, exit	*/
			/* no need to keep it !			*/
			/* -----------------------------------	*/

 if ((int)fscore < (int)(p_ope->gen_kept[p_ope->max_gen_kept-1].score)) 
    return;


			/* -----------------------------------	*/
			/* else put in the right place		*/
			/* -----------------------------------	*/
 fini = 0;

 for (i = 0; i < p_ope->max_gen_kept && ! fini; i++) {

    if ((int)fscore >= (int)(p_ope->gen_kept[i].score)) {


			/* -----------------------------------	*/
			/* if last gene is kept, decrements	*/
			/* the set the gene as kept		*/
			/* -----------------------------------	*/

       j = p_ope->max_gen_kept - 1;
       if (p_ope->gen_kept[j].gen_ptr != NULL
		&& p_ope->gen_kept[j].gen_ptr->kept_by > 0) {

	  p_ope->gen_kept[j].gen_ptr->kept_by--;

#if 0	/* DEBUG */
	  (void)fprintf(stdout, ">>>gene %s now kept=%d\n",
		p_ope->gen_kept[j].gen_ptr->fseqptr->name, 
		p_ope->gen_kept[j].gen_ptr->kept_by);
#endif
       }

			/* -----------------------------------	*/
			/* translate the kept genes one		*/
			/* position right			*/
			/* -----------------------------------	*/

       for (j = p_ope->max_gen_kept - 1; j > i; j--) {
	  p_ope->gen_kept[j].gen_ptr = p_ope->gen_kept[j-1].gen_ptr;
	  p_ope->gen_kept[j].score = p_ope->gen_kept[j-1].score;
       }

       p_ope->gen_kept[i].gen_ptr = p_gen;
       p_ope->gen_kept[i].score = (SCORE)fscore;
			/* set the gene as kept	*/
       p_ope->gen_kept[i].gen_ptr->kept_by++;
       p_ope->actual_gen_kept  = MIN(p_ope->max_gen_kept, p_ope->actual_gen_kept + 1);
       fini = 1;
    }

 }

#if 0		/* debug	*/
 WriteKeptGene(p_ope, len_thresh, 10000, 0.001, 0, 0, stderr);
#endif

 return ;

}

/* -------------------------------------------- */
/* KeepOpeInGeneF: keep the operon ope in the	*/
/* gen_kept array				*/
/* -------------------------------------------- */
void KeepOpeInGeneF(GenomeSeqPtr p_gen, OperonSeqPtr ope_ptr, float fscore)
{

 int i, j, fini;

			/* -----------------------------------	*/
			/* if fscore is inf. to min, exit	*/
			/* no need to keep it !			*/
			/* -----------------------------------	*/

  if ((int)fscore < (int)(p_gen->ope_kept[p_gen->max_ope_kept-1].score))
     return ;


			/* -----------------------------------	*/
			/* else put in the right place		*/
			/* -----------------------------------	*/
 fini = 0;

 for (i = 0; i < p_gen->max_ope_kept && ! fini; i++) {

    if ((int)fscore >= (int)(p_gen->ope_kept[i].score)) {


			/* -----------------------------------	*/
			/* if last ope is kept, decrements	*/
			/* the set the gene as kept		*/
			/* -----------------------------------	*/

       j = p_gen->max_ope_kept - 1;
       if (p_gen->ope_kept[j].ope_ptr != NULL
		&& p_gen->ope_kept[j].ope_ptr->kept_by > 0) {

	  p_gen->ope_kept[j].ope_ptr->kept_by--;

#if 0	/* DEBUG */
	  (void)fprintf(stdout, ">>>gene %s now kept=%d\n",
		p_gen->ope_kept[j].ope_ptr->fseqptr->name, 
		p_gen->ope_kept[j].ope_ptr->kept_by);
#endif
       }

			/* -----------------------------------	*/
			/* translate the kept genes one		*/
			/* position right			*/
			/* -----------------------------------	*/

       for (j = p_gen->max_ope_kept - 1; j > i; j--) {
	  p_gen->ope_kept[j].ope_ptr = p_gen->ope_kept[j-1].ope_ptr;
	  p_gen->ope_kept[j].score = p_gen->ope_kept[j-1].score;
       }

       p_gen->ope_kept[i].ope_ptr = ope_ptr;
       p_gen->ope_kept[i].score = (SCORE)fscore;
			/* set the operon gene as kept	*/
       p_gen->ope_kept[i].ope_ptr->kept_by++;
       p_gen->actual_ope_kept  = MIN(p_gen->max_ope_kept, p_gen->actual_ope_kept + 1);
       fini = 1;
    }

 }


   return ;
}


/* -------------------------------------------- */
/* return index of the first element in		*/
/* p_gen->ope_kept array which have a		*/
/* score <= MAX					*/
/* returns -1 if not found			*/
/* -------------------------------------------- */
int FindInfScoreInOpeKept(GenomeSeqPtr p_gen, int max)
{
   int i;

   for (i = 0; i < p_gen->actual_ope_kept ; i++)
      if ((int)(p_gen->ope_kept[i].score) <= max)
	return i;

   return -1;

}


/* -------------------------------------------- */
/* IsGenInOpeKeptR: return the score of gene	*/
/* gen_ptr if gen_ptr is in the			*/
/* p_ope->gen_kept[i] table and if its score	*/
/* is > min, else returns -1			*/
/* -------------------------------------------- */
int IsGenInOpeKeptR(OperonSeqPtr p_ope, GenomeSeqPtr p_gen, int min)
{
   int i;

   for (i = 0; i < p_ope->actual_gen_kept; i++)
      if (p_ope->gen_kept[i].gen_ptr == p_gen)
	return (((int)(p_ope->gen_kept[i].score) > min) ?
		(int)(p_ope->gen_kept[i].score) :
		-1);

   return -1;

}

/* -------------------------------------------------------------------- */
/* ComputeStatForOperon: compute simple stat for kept gene scores for	*/
/* the operon ope_ptr.							*/
/* Simply count the number of score above scores in aleascore array	*/
/* after sorting it							*/
/* -------------------------------------------------------------------- */
void ComputeStatForOperon(OperonSeqPtr ope_ptr, int *aleascore, int n_alea)
{
   int i, j, n;

#define DEBUGGG

#ifdef DEBUGGG
     fprintf(stderr, "\n");
#endif

   for (i = 0; i < ope_ptr->actual_gen_kept; i++) {
      n = 0;
      for (j = 0; j < n_alea; j++) {
	 if ((int)(ope_ptr->gen_kept[i].score) <= aleascore[j]) {
	    n++;
	 }
      }
      ope_ptr->gen_kept[i].stat = (float)n/(float)n_alea;

#ifdef DEBUGGG
      fprintf(stderr, "ope %s gen %s score %4d prob %3e\n",
	ope_ptr->mnemo, ope_ptr->gen_kept[i].gen_ptr->mnemo,
	(int)(ope_ptr->gen_kept[i].score), ope_ptr->gen_kept[i].stat);
#endif

   }
   
   return;
}

/* -------------------------------------------- */
/* KeepOpeInGeneB: keep the operon ope in the	*/
/* gen_kept array, the operon ope MUST be 	*/
/* already present (with a fasta score) 	*/
/* -------------------------------------------- */
void KeepOpeInGeneB(GenomeSeqPtr p_gen, OperonSeqPtr ope_ptr,
			float fscore, float stat)
{

 int i, j, fini, rempl;

			/* -----------------------------------	*/
			/* if fscore is inf. to min, exit	*/
			/* no need to keep it !			*/
			/* -----------------------------------	*/

  if ((int)fscore < (int)(p_gen->ope_kept[p_gen->max_ope_kept-1].score)) 
     return;


			/* -----------------------------------	*/
			/* else put in the right place		*/
			/* -----------------------------------	*/
 fini = 0;
 rempl = -1;

 for (i = 0; i < p_gen->max_ope_kept && ! fini; i++) {

    if ((rempl == -1) && ((int)fscore >= (int)(p_gen->ope_kept[i].score))) {
       rempl = i;
    }

    if (p_gen->ope_kept[i].ope_ptr == ope_ptr) {

			/* should never append			*/
       if (rempl == -1)
	   MyError(stderr, 5, ERROR_TAG, 
	      "KeepOpeInGeneB: error: rempl=-1 : should never happen !!!");
       

			/* this is the old place		*/
	

			/* -----------------------------------	*/
			/* we must shit on the right all  	*/
			/* elements between rempl and i		*/
			/* Note: i > rempl			*/
			/* -----------------------------------	*/

			/* -----------------------------------	*/
			/* as we replace the same ope_ptr	*/
			/* no need to re-evaluate the counter	*/
			/* ope_ptr->kept_by			*/
			/* -----------------------------------	*/
			/* translate the kept genes one		*/
			/* position right			*/
			/* -----------------------------------	*/
       if (stat < -0.1)			/* keep stat */
	  stat = p_gen->ope_kept[i].stat;

       for (j = i; j > rempl; j--) {
	  p_gen->ope_kept[j].ope_ptr = p_gen->ope_kept[j-1].ope_ptr;
	  p_gen->ope_kept[j].score = p_gen->ope_kept[j-1].score;
	  p_gen->ope_kept[j].stat = p_gen->ope_kept[j-1].stat;
       }

       p_gen->ope_kept[rempl].ope_ptr = ope_ptr;
       p_gen->ope_kept[rempl].score = (SCORE)fscore;
       p_gen->ope_kept[rempl].stat = stat;
       fini = 1;
    }

 }


 return ;

}

/* -------------------------------------------- */
/* ResetAllScores: correct the score of operon	*/
/* and associated homolog by correction		*/
/* -------------------------------------------- */
void ResetAllScores(OperonSeqPtr *ope, int nb_ope,
		int corr)
{

   int i,j;

   for (i = 0; i < nb_ope; i++) {
      if (ope[i]->homolog.gen_ptr != NOT_G)
	 ope[i]->homolog.score = (SCORE)((int)ope[i]->homolog.score + corr);
      for (j = 0; j < ope[i]->actual_gen_kept; j++)
	 ope[i]->gen_kept[j].score = (SCORE)(
		   (int)ope[i]->gen_kept[j].score + corr);
   }

   return;

}

/* -------------------------------------------- */
/* set the homolog field of all genomic genes	*/
/* to NOT_O if it is NULL			*/
/* -------------------------------------------- */
void set_NULL2NOT_O_Gene(GenomeSeqPtr *gen, int nb_gen)
{
   int i;

   for (i = 0; i < nb_gen; i++) 
      if (gen[i]->homol_ptr == NULL)
	 gen[i]->homol_ptr = NOT_O;

   return;
}

/* -------------------------------------------- */
/* set the homolog field of all operon genes	*/
/* Note : if bijectif, scores of genomic genes	*/
/* with operon must have been computed		*/
/* Note:					*/
/* seuil_score must be incremented before by	*/
/* BASE_BEST as we compare bestfit scores	*/
/* -------------------------------------------- */
void SetAllOpeHomologs(OperonSeqPtr *p_ope, int nb_ope,
		 float len_thresh, int seuil_score,
		 int bijectif, int verbose)
{

   int i;


			/* first run	*/
   for (i = 0; i < nb_ope; i++) {

      if(verbose)
	 (void)fprintf(stderr, "\r" COMPLETION_TAG " sequence %d / %d\t\t", i+1, nb_ope);

      SetOneHomolog(p_ope[i], len_thresh, seuil_score, bijectif);

   }
#if RECURS_HOMOLOG
   SetAllOpeLastHomologs(p_ope, nb_ope, len_thresh, seuil_score,
			bijectif, verbose);
#endif
   return;

}

/* -------------------------------------------- */
/* set the homolog field of all operon genes    */
/* SECOND PASS after SetAllOpeHomologs		*/
/* reason: in the first pass, one cannot 	*/
/* decide in a situation where:			*/
/* (or risk to miss the ope1-gen1 homology	*/
/* OPE    GEN_KEPT OPE_KEPT_OF_GEN_KEPT		*/
/* ope1 -> gen1 -> ope3				*/
/*      -> gen1 -> ope1				*/
/* i.e.						*/
/* ope3 has a best score than ope1 with gen1	*/
/* but ope3 has another homolog, but not yet	*/
/* found, we must do a second run (ope3 will	*/
/* be assigned its homolog in the first run),	*/
/* and then ope1 will be the best "free" hit	*/
/* in ope_kept of gen1, and they will be set	*/
/* as homologs					*/
/* -------------------------------------------- */
void SetAllOpeLastHomologs(OperonSeqPtr *p_ope, int nb_ope,
			float len_thresh, int seuil_score,
			int bijectif, int verbose)
{
   int i;

			/* second run	*/
   for (i = 0; i < nb_ope; i++) {

      if(verbose)
	 (void)fprintf(stderr, "\r" COMPLETION_TAG " sequence %d / %d\t\t", i+1, nb_ope);

      sSetOneLastHomolog(p_ope[i], len_thresh, seuil_score, bijectif);

   }

   return;

}


/* -------------------------------------------- */
/* set the homolog field of one operon genes	*/
/* Note : if bijectif, scores of genomic genes	*/
/* with operon must have been computed		*/
/* Note:					*/
/* seuil_score must be incremented before by	*/
/* BASE_BEST as we compare bestfit scores	*/
/* -------------------------------------------- */
void SetOneHomolog(OperonSeqPtr ope, float len_thresh,
		int seuil_score, int bijectif)
{
   SCORE limscore, limjscore;
   int	found, i, j;

   found = 0;

   if (ope->actual_gen_kept > 0 && (int)(ope->gen_kept[0].score) >= seuil_score) {
      limscore = ope->gen_kept[0].score;
      i = 0;
      found = 0;
				/* if several scores are ex-aequo	*/
				/* find the first gene whose best	*/
				/* operon hit is the operon ope		*/
      while ( ! found && i < ope->actual_gen_kept &&
	     (int)(ope->gen_kept[i].score) >= (int)limscore ) {
	 if (len_thresh < 0.0 || LENRATIO(ope->fseqptr->length, 
		 ope->gen_kept[i].gen_ptr->fseqptr->length) <= len_thresh) {
				/* ------------------------------------	*/
				/* if no bijective relation is required	*/
				/* take the first (best) ope->gen_kept	*/
				/* ------------------------------------	*/
	    if (!bijectif) {
	       ope->homolog.gen_ptr = ope->gen_kept[i].gen_ptr;
	       /* ope->homolog.gen_ptr->homolog becomes NOT_O in the 	*/
	       /* no bijective case					*/
	       ope->homolog.score = ope->gen_kept[i].score;
	       ope->homolog.stat = ope->gen_kept[i].stat;
				/* ------------------------------------	*/
				/* if no bijectif, no meaning of	*/
				/* gen_ptr->homol_ptr, put it to the	*/
				/* last "homolog" ope			*/
				/* ------------------------------------	*/
	       ope->homolog.gen_ptr->homol_ptr = ope;
	       found = 1;
	    }
	    else {
				/* ------------------------------------	*/
				/* if bijective relation is required	*/
				/* for each ope_kept[j] of best		*/
				/* ope->gen_kept[i], take the first	*/
				/* which has ope as a best (i.e. > 	*/
				/* limscore) hit			*/
				/* ------------------------------------	*/

				/* ------------------------------------	*/
				/* if gen_kept[i] has not yet homolog	*/
				/* ------------------------------------	*/
	       if (ope->gen_kept[i].gen_ptr->homol_ptr == NULL) {
				
				/* ------------------------------------	*/
				/* while ope_kept[j] of gen_kept[i]	*/
				/* has an homolog, continue to see if	*/
				/* ope could be the next best of 	*/
				/* gen_kept[i]				*/
				/* ------------------------------------	*/
		  j = 0;
		  limjscore = ope->gen_kept[i].gen_ptr->ope_kept[0].score;
				/* ------------------------------------	*/
				/* if ope and gen_kept[i] have a score	*/
				/* inf to best score of gen_kept[i], we	*/
				/* must not take them as homologs	*/
				/* ------------------------------------	*/
		  if (limscore >= limjscore) {

		     while( j < ope->gen_kept[i].gen_ptr->actual_ope_kept &&
			  ope->gen_kept[i].gen_ptr->ope_kept[j].ope_ptr->homolog.gen_ptr != NULL &&
	       		  ope->gen_kept[i].gen_ptr->ope_kept[j].score >= limjscore)
				j++;


				/* ------------------------------------	*/
				/* now, if j < kept ope of gen_kept[i]	*/
				/* we are on the first ope_kept[j]	*/
				/* with no homolog, we must verify	*/
				/* that it is ope			*/
				/* ------------------------------------	*/
		     if (j < ope->gen_kept[i].gen_ptr->actual_ope_kept &&
	       		  ope->gen_kept[i].gen_ptr->ope_kept[j].score >= limjscore &&
			  ope == ope->gen_kept[i].gen_ptr->ope_kept[j].ope_ptr) {
				/* ------------------------------------	*/
				/* if the best similar operon of	*/
				/* homolog has not already an homolog	*/
				/* and if the best similar operon of	*/
				/* homolog is ope, then these are two	*/
				/* "homologs"				*/
				/* Note:				*/
				/* ope->homolog.gen_ptr->ope_kept[i]	*/
				/* must exists as the score is >=	*/
				/* seuil_score				*/
				/* ------------------------------------	*/
				/* Attention ici on doit faire atten-	*/
				/* tion que le gen en question ne soit	*/
				/* pas deja affecte a un meilleur ope	*/
				/* au niveau score			*/
				/* ------------------------------------	*/

	               ope->homolog.gen_ptr = ope->gen_kept[i].gen_ptr;
	               ope->homolog.score = ope->gen_kept[i].score;
	               ope->homolog.stat = ope->gen_kept[i].stat;
		       ope->homolog.gen_ptr->homol_ptr = ope;
	               found = 1;
		     }
		  }
	       }
	    }
	 }
	 i++;
      }
   }

   if (! found ) {
	 ope->homolog.gen_ptr = NOT_G;
	 ope->homolog.score   = (int)0;
	 ope->homolog.stat   = -1.0;
   }

   return;

}

/* -------------------------------------------- */
/* set the homolog field of one operon genes	*/
/* Note : a first pass must have done and	*/
/* this routine examines only ope which have	*/
/* NOT homolog					*/
/* ALL ope must have an homolog or a NOT	*/
/* homolog field				*/
/* if bijectif, scores of genomic genes		*/
/* with operon must have been computed		*/
/* Note:					*/
/* seuil_score must be incremented before by	*/
/* BASE_BEST as we compare bestfit scores	*/
/* -------------------------------------------- */
static void sSetOneLastHomolog(OperonSeqPtr ope, float len_thresh,
		int seuil_score, int bijectif)
{
   SCORE limscore;
   int	found, i, j;

   if (ope->homolog.gen_ptr != NOT_G)
	return;

   found = 0;

   if (ope->actual_gen_kept > 0 && (int)(ope->gen_kept[0].score) >= seuil_score) {
      limscore = ope->gen_kept[0].score;
      i = 0;
      found = 0;
				/* if several scores are ex-aequo	*/
				/* find the first gene whose best	*/
				/* operon hit is the operon ope		*/
      while ( ! found && i < ope->actual_gen_kept &&
	     (int)(ope->gen_kept[i].score) >= (int)limscore ) {
	 if (len_thresh < 0.0 || LENRATIO(ope->fseqptr->length, 
		 ope->gen_kept[i].gen_ptr->fseqptr->length) <= len_thresh) {
				/* ------------------------------------	*/
				/* if no bijective relation is required	*/
				/* take the first (best) ope->gen_kept	*/
				/* ------------------------------------	*/
	    if (!bijectif) {
	       ope->homolog.gen_ptr = ope->gen_kept[i].gen_ptr;
	       /* ope->homolog.gen_ptr->homolog becomes NOT_O in the 	*/
	       /* no bijective case					*/
	       ope->homolog.score = ope->gen_kept[i].score;
	       ope->homolog.stat = ope->gen_kept[i].stat;
				/* ------------------------------------	*/
				/* if no bijectif, no meaning of	*/
				/* gen_ptr->homol_ptr, put it to the	*/
				/* last "homolog" ope			*/
				/* ------------------------------------	*/
	       ope->homolog.gen_ptr->homol_ptr = ope;
	       found = 1;
	    }
	    else {
				/* ------------------------------------	*/
				/* if bijective relation is required	*/
				/* for each ope_kept[j] of best		*/
				/* ope->gen_kept[i], take the first	*/
				/* which has ope as a best (i.e. > 	*/
				/* limscore) hit			*/
				/* ------------------------------------	*/

				/* ------------------------------------	*/
				/* if gen_kept[i] has not yet homolog	*/
				/* ------------------------------------	*/
	       if (ope->gen_kept[i].gen_ptr->homol_ptr == NULL) {
				
				/* ------------------------------------	*/
				/* while ope_kept[j] of gen_kept[i]	*/
				/* has an homolog, continue to see if	*/
				/* ope could be the next best of 	*/
				/* gen_kept[i]				*/
				/* ------------------------------------	*/
		  j = 0;
				/* ------------------------------------	*/
				/* if ope and gen_kept[i] have a score	*/
				/* inf to best score of gen_kept[i], we	*/
				/* must not take them as homologs	*/
				/* ------------------------------------	*/

		  while( j < ope->gen_kept[i].gen_ptr->actual_ope_kept &&
			  ope->gen_kept[i].gen_ptr->ope_kept[j].ope_ptr->homolog.gen_ptr != NOT_G &&
	       		  ope->gen_kept[i].gen_ptr->ope_kept[j].score >= limscore)
				j++;


				/* ------------------------------------	*/
				/* now, if j < kept ope of gen_kept[i]	*/
				/* we are on the first ope_kept[j]	*/
				/* with no homolog, we must verify	*/
				/* that it is ope			*/
				/* ------------------------------------	*/
		     if (j < ope->gen_kept[i].gen_ptr->actual_ope_kept &&
	       		  ope->gen_kept[i].gen_ptr->ope_kept[j].score >= limscore &&
			  ope == ope->gen_kept[i].gen_ptr->ope_kept[j].ope_ptr) {
				/* ------------------------------------	*/
				/* if the best similar operon of	*/
				/* homolog has not already an homolog	*/
				/* and if the best similar operon of	*/
				/* homolog is ope, then these are two	*/
				/* "homologs"				*/
				/* Note:				*/
				/* ope->homolog.gen_ptr->ope_kept[i]	*/
				/* must exists as the score is >=	*/
				/* seuil_score				*/
				/* ------------------------------------	*/
				/* Attention ici on doit faire atten-	*/
				/* tion que le gen en question ne soit	*/
				/* pas deja affecte a un meilleur ope	*/
				/* au niveau score			*/
				/* ------------------------------------	*/

	               ope->homolog.gen_ptr = ope->gen_kept[i].gen_ptr;
	               ope->homolog.score = ope->gen_kept[i].score;
	               ope->homolog.stat = ope->gen_kept[i].stat;
		       ope->homolog.gen_ptr->homol_ptr = ope;
	               found = 1;
		     }
	          }
	    }
	 }
	 i++;
      }
   }

   if (! found ) {
	 ope->homolog.gen_ptr = NOT_G;
	 ope->homolog.score   = (int)0;
	 ope->homolog.stat   = -1.0;
   }

   return;

}

/* -------------------------------------------- */
/* set the homolog field of one operon genes	*/
/* Note : if bijectif, scores of genomic genes	*/
/* with operon must have been computed		*/
/* Note:					*/
/* seuil_score must be incremented before by	*/
/* BASE_BEST as we compare bestfit scores	*/
/* -------------------------------------------- */
void SetOneHomologOld(OperonSeqPtr ope, float len_thresh,
		int seuil_score, int bijectif)
{
   SCORE limscore;
   int	found, i, j;

   found = 0;

   if (ope->actual_gen_kept > 0 && (int)(ope->gen_kept[0].score) >= seuil_score) {
      limscore = ope->gen_kept[0].score;
      i = 0;
      found = 0;
				/* if several scores are ex-aequo	*/
				/* find the first gene whose best	*/
				/* operon hit is the operon ope		*/
      while ( ! found && i < ope->actual_gen_kept &&
	     (int)(ope->gen_kept[i].score) >= (int)limscore ) {
	 if (len_thresh < 0.0 || LENRATIO(ope->fseqptr->length, 
		 ope->gen_kept[i].gen_ptr->fseqptr->length) <= len_thresh) {
				/* ------------------------------------	*/
				/* if no bijective relation is required	*/
				/* take the first (best) ope->gen_kept	*/
				/* ------------------------------------	*/
	    if (!bijectif) {
	       ope->homolog.gen_ptr = ope->gen_kept[i].gen_ptr;
	       /* ope->homolog.gen_ptr->homolog becomes NOT_O in the 	*/
	       /* no bijective case					*/
	       ope->homolog.score = ope->gen_kept[i].score;
	       ope->homolog.stat = ope->gen_kept[i].stat;
				/* ------------------------------------	*/
				/* if no bijectif, no meaning of	*/
				/* gen_ptr->homol_ptr, put it to the	*/
				/* last "homolog" ope			*/
				/* ------------------------------------	*/
	       ope->homolog.gen_ptr->homol_ptr = ope;
	       found = 1;
	    }
	    else {
				/* ------------------------------------	*/
				/* if bijective relation is required	*/
				/* for each ope_kept[j] of best		*/
				/* ope->gen_kept[i], take the first	*/
				/* which has ope as a best (i.e. > 	*/
				/* limscore) hit			*/
				/* ------------------------------------	*/

				/* ------------------------------------	*/
				/* if gen_kept[i] has not yet homolog */
				/* ------------------------------------	*/
	          if (ope->gen_kept[i].gen_ptr->homol_ptr == NULL) {
				
				/* ------------------------------------	*/
				/* while ope_kept[j] of gen_kept[i]	*/
				/* has an homolog, continue to see if	*/
				/* ope could be the next best of 	*/
				/* gen_kept[i]				*/
				/* ------------------------------------	*/
		     j = 0;
		     while( j < ope->gen_kept[i].gen_ptr->actual_ope_kept &&
			  /* ope->gen_kept[i].gen_ptr->ope_kept[j].ope_ptr->homolog && */
			  ope->gen_kept[i].gen_ptr->ope_kept[j].ope_ptr->homolog.gen_ptr != NULL &&
	       		  ope->gen_kept[i].gen_ptr->ope_kept[j].score >= limscore)
				j++;


				/* ------------------------------------	*/
				/* now, if j < kept ope of gen_kept[i]	*/
				/* we are on the first ope_kept[j]	*/
				/* with no homolog, we must verify	*/
				/* that it is ope			*/
				/* ------------------------------------	*/
		     if (j < ope->gen_kept[i].gen_ptr->actual_ope_kept &&
			  ope->gen_kept[i].gen_ptr->ope_kept[j].score >= limscore &&
			  ope == ope->gen_kept[i].gen_ptr->ope_kept[j].ope_ptr) {
				/* ------------------------------------	*/
				/* if the best similar operon of	*/
				/* homolog has not already an homolog	*/
				/* and if the best similar operon of	*/
				/* homolog is ope, then these are two	*/
				/* "homologs"				*/
				/* Note:				*/
				/* ope->homolog.gen_ptr->ope_kept[i]	*/
				/* must exists as the score is >=	*/
				/* seuil_score				*/
				/* ------------------------------------	*/
				/* Attention ici on doit faire atten-	*/
				/* tion que le gen en question ne soit	*/
				/* pas deja affecte a un meilleur ope	*/
				/* au niveau score			*/
				/* ------------------------------------	*/

	               ope->homolog.gen_ptr = ope->gen_kept[i].gen_ptr;
	               ope->homolog.score = ope->gen_kept[i].score;
	               ope->homolog.stat = ope->gen_kept[i].stat;
		       ope->homolog.gen_ptr->homol_ptr = ope;
	               found = 1;
		     }
	       }
#if OLD_VERSION
	       j = 0;
	       cont = 1;
	       while ( cont == 1 && j < ope->gen_kept[i].gen_ptr->actual_ope_kept &&
		       ope->gen_kept[i].gen_ptr->ope_kept[j].score >= limscore) {
				/* ------------------------------------	*/
				/* if gen_kept[i] has not yet homolog */
				/* ------------------------------------	*/
	          if (ope->gen_kept[i].gen_ptr->homol_ptr == NULL) {
				/* ------------------------------------	*/
				/* if ope is the first preferred of	*/
				/* gen_kept[i] which has no homolog	*/
				/* ------------------------------------	*/
		     if (ope == ope->gen_kept[i].gen_ptr->ope_kept[j].ope_ptr) {
				/* ------------------------------------	*/
				/* if the best similar operon of	*/
				/* homolog has not already an homolog	*/
				/* and if the best similar operon of	*/
				/* homolog is ope, then these are two	*/
				/* "homologs"				*/
				/* Note:				*/
				/* ope->homolog.gen_ptr->ope_kept[i]	*/
				/* must exists as the score is >=	*/
				/* seuil_score				*/
				/* ------------------------------------	*/
				/* Attention ici on doit faire atten-	*/
				/* tion que le gen en question ne soit	*/
				/* pas deja affecte a un meilleur ope	*/
				/* au niveau score			*/
				/* ------------------------------------	*/

	               ope->homolog.gen_ptr = ope->gen_kept[i].gen_ptr;
	               ope->homolog.score = ope->gen_kept[i].score;
	               ope->homolog.stat = ope->gen_kept[i].stat;
		       ope->homolog.gen_ptr->homol_ptr = ope;
	               found = 1;
		     }
				/* ------------------------------------	*/
				/* here, gen_kept[i] has not yet an	*/
				/* homolog, but ope is not its best	*/
				/* hit, so stops searching in the	*/
				/* gen_kept table			*/
				/* ------------------------------------	*/
		     cont = 0;
		  }
		  j++;
	       }
#endif
	    }
	 }
	 i++;
      }
   }

   if (! found ) {
	 ope->homolog.gen_ptr = NOT_G;
	 ope->homolog.score   = (int)0;
	 ope->homolog.stat   = -1.0;
   }

   return;

}

/* -------------------------------------------- */
/* ComparBestFitScoreGenKept: compares bestfit	*/
/* scores of GenKept gen 1 and gen2		*/
/* struct for use in qsort			*/
/* return 0: if gen1.score == gen2.score	*/
/*       +1: if gen1.score <  gen2.score	*/
/*       -1: if gen1.score >  gen2.score	*/
/* Note: the score must be sorted in descending	*/
/* order					*/
/* -------------------------------------------- */
int ComparBestFitScoreGenKept(const void *gen1, const void *gen2)
{
   if ((int)(((GenKeptPtr)gen1)->score) <
       (int)(((GenKeptPtr)gen2)->score)   ) 

      return +1;

   else if ((int)(((GenKeptPtr)gen1)->score) >
	    (int)(((GenKeptPtr)gen2)->score)  )

     return -1;

   else

     return 0;

}

/* -------------------------------------------- */
/* WriteKeptGene: write to fout the gene name	*/
/* kept with operon ope with their score	*/
/* and stat if stat_on				*/
/* if their relative length agree len_thresh 	*/
/* if short_output, the output is shortened	*/
/* -------------------------------------------- */
void WriteKeptGene(OperonSeqPtr p_ope, float len_thresh,
			int correction_score, float normscore,
			int stat_on, int short_output, FILE *fout)
{
  int i;
  float len_ratio;

#define PPCOMPL (void)fprintf(fout, COMPLETION_TAG

  if (! short_output) {
     (void)fprintf(fout,"\n");
     PPCOMPL "List of similar genes with operon %s:\n",
        p_ope->mnemo);
  }

  for (i = 0; i < p_ope->actual_gen_kept; i++) {

			/* si limite de taille,		*/
			/* on n'imprime pas		*/

	len_ratio = LENRATIO(p_ope->fseqptr->length,
			p_ope->gen_kept[i].gen_ptr->fseqptr->length);

	if (! (len_thresh > 0) || len_ratio <= len_thresh) {

           if (! short_output) {
             PPCOMPL "\tgene %s (score %6.2f ",
		p_ope->gen_kept[i].gen_ptr->fseqptr->name,
		(((float)(p_ope->gen_kept[i].score) -
		      (float)correction_score)*normscore));
	     if (stat_on)
		(void)fprintf(fout,"prob %.3e ", p_ope->gen_kept[i].stat);
             (void)fprintf(fout,"dlen ratio %5.2f)\n", len_ratio);
	   }
	   else {
	     (void)fprintf(fout,"%s %s %6.2f ",
			p_ope->mnemo,
			p_ope->gen_kept[i].gen_ptr->fseqptr->name,
			(((float)(p_ope->gen_kept[i].score) -
			(float)correction_score)*normscore));
	     if (stat_on)
		(void)fprintf(fout,"prob %.3e ", p_ope->gen_kept[i].stat);
             (void)fprintf(fout,"%5.2f %4d\n", len_ratio, p_ope->actual_gen_kept);
	   }

	}

  }

  return;

#undef PPCOMPL

}

/* -------------------------------------------- */
/* GenomeSeqParseInfo : parse the information 	*/
/* in the comment of the FastaSequence		*/
/* if standard_fasta_format == 1, the sequence	*/
/* is in standard fasta format, else it is	*/
/* in "mnemo format" (see README.opscan)	*/
/* -------------------------------------------- */

static Bool GenomeSeqParseInfo(GenomeSeqPtr p_gen, int standard_fasta_format) 
{

   static int	counter = 0, deja_erreur = 0;
   char		*pos1,*pos2, tmpname[MNEMOCHAR];

   if (p_gen == NULL)
       return Faux;

   if ((int)(strlen(p_gen->fseqptr->name)) >= MNEMOCHAR) {

        MyError(stderr, 0, ERROR_TAG, 
	  "GenomeSeqParseInfo:: error name %s is longer than MNEMOCHAR (%d)\n",
		p_gen->fseqptr->name, MNEMOCHAR);

   }
   else {

        (void)strcpy(p_gen->mnemo, p_gen->fseqptr->name);

   }
   
   counter++;
   if (standard_fasta_format == 0) {
     (void)strncpy(tmpname, p_gen->fseqptr->name, GEN_NAME_LEN-1);
     tmpname[GEN_NAME_LEN-1]='\0';
     pos1 = strchr(tmpname,(int)'@');
     pos2 = pos1 ? strchr(pos1+1,(int)'@') : NULL;
 
     
     if (pos1 && pos2) {
			/* there is a least two '@'					*/
			/* name is a complex name i.e. <gene_ID>@<org_ID>@<contig_ID>	*/
       *pos1 = '\0';
       *pos2 = '\0';
       (void)strcpy(p_gen->gene_name, tmpname);
       (void)strncpy(p_gen->org_name, pos1, ORG_NAME_LEN-1);
       p_gen->org_name[ORG_NAME_LEN-1] = '\0';
       (void)strncpy(p_gen->contig_ID, pos2, CTG_NAME_LEN-1);
       p_gen->contig_ID[CTG_NAME_LEN-1] = '\0';

     }
     else {
	 if (deja_erreur == 0) {
	    MyError(stderr, 0, REMARK_TAG, 
	       "GenomeSeqParseInfo:: genomic sequence mnemo (number %d) is not in the correct format", counter);
	    MyError(stderr, 0, REMARK_TAG, 
	      "      it is: %s", p_gen->fseqptr->name);
	    MyError(stderr, 0, REMARK_TAG, 
	      "      and it should be: <gene_ID>@<org_ID>@<contig_ID> ...");
	    MyError(stderr, 0, REMARK_TAG, 
	      "      The name %s only will be used", p_gen->fseqptr->name);
	    MyError(stderr, 0, REMARK_TAG,
	       "      or use the standard fasta format option for genomic sequences database");
 	    MyError(stderr, 0, REMARK_TAG,
	       "      (see the README.opscan file)");
	    MyError(stderr, 0, REMARK_TAG,
	      "      Nevertheless Continue reading assuming standard fasta format");
	    deja_erreur = 1;
	 }
         (void)strncpy(p_gen->gene_name, p_gen->fseqptr->name,GEN_NAME_LEN-1);
         p_gen->gene_name[GEN_NAME_LEN-1]='\0';
         (void)strcpy(p_gen->org_name, "dummy");
         (void)strcpy(p_gen->contig_ID, "dummy");
     }

     if (sscanf(p_gen->fseqptr->comment,"%d %d %d %c\n",
			&(p_gen->rel_number),
			&(p_gen->from), &(p_gen->to),
			&(p_gen->strand)) != 4) {
        if (deja_erreur < 2) {
	   MyError(stderr, 0, REMARK_TAG,
	     "GenomeSeqParseInfo:: genomic sequence number, from, to and strand are not in the correct format");
	   MyError(stderr, 0, REMARK_TAG,
	     "      it is: (%s)", p_gen->fseqptr->comment);
	   MyError(stderr, 0, REMARK_TAG,
	     "      and it should be:  rel_number from to strand");
	   MyError(stderr, 0, REMARK_TAG,
	     "      or use the standard fasta format option for genomic sequences database");
  	   MyError(stderr, 0, REMARK_TAG,
	     "      (see the README.opscan file)");
	   MyError(stderr, 0, REMARK_TAG,
	     "      Nevertheless Continue reading assuming standard fasta format");
	}
	p_gen->rel_number = counter;
	deja_erreur = 3;
    }

#if 0	/* debug	*/
    (void)fprintf(stdout, "%s:rel_number:%d\n", p_gen->fseqptr->name, p_gen->rel_number);
#endif

   }
   else {
			/* standard_fasta_format == 1		*/
			/* gene name will be the fasta name	*/
       (void)strncpy(p_gen->gene_name, p_gen->fseqptr->name,GEN_NAME_LEN-1);
       p_gen->gene_name[GEN_NAME_LEN-1]='\0';
			/* dummy value	*/
       (void)strcpy(p_gen->org_name, "dummy");
       (void)strcpy(p_gen->contig_ID, "dummy");
       p_gen->rel_number = counter;
       p_gen->from   = -1;
       p_gen->to     = -1;
       p_gen->strand = UNK_STRAND;
   }

   return Vrai;

} 


/* -------------------------------------------- */
/* allocation d'un GenomeSequence              */
/* -------------------------------------------- */

GenomeSeqPtr NewGenomeSequence(int ope_kept)
{

   GenomeSeqPtr p_gen;
   int		i;

   if (! (p_gen = NEW(GenomeSequence)))
      return NULL;

   if (! (p_gen->fseqptr = NewFastaSequence())) {
      MyError(stderr,0, ERROR_TAG,
	"NewGenomeSequence:: cannot create NewFastaSequence\n");
      free(p_gen);
      return NULL;
   }

   if (! (p_gen->ope_kept = NEWN(OpeKept, ope_kept)) ) {
      MyError(stderr,0, ERROR_TAG,
	"NewGenomeSequence:: cannot create ope_kept\n");
      free(p_gen->fseqptr);
      free(p_gen);
      return NULL;
   }

   p_gen->from = UNK_IGNORE;
   p_gen->to   = UNK_IGNORE;

   p_gen->kept_by = NO_KEPT;
   p_gen->class = NO_CLASS;

   p_gen->max_ope_kept      = ope_kept;
   p_gen->actual_ope_kept   = 0;
   p_gen->bestfitted        = 0;

   for (i = 0; i < p_gen->max_ope_kept; i++) {
      p_gen->ope_kept[i].ope_ptr = NULL;
      p_gen->ope_kept[i].score = (int)0;
      p_gen->ope_kept[i].stat = -1.0;
   }

   p_gen->mnemo[0]       = '\0';
   p_gen->rel_number     = -1;
   p_gen->contig_ID[0]   = '\0';
   p_gen->org_name[0]    = '\0';
   p_gen->gene_name[0]   = '\0';

   p_gen->homol_ptr      = NULL;

   p_gen->strand         = UNK_STRAND;

   return p_gen;
}

/* -------------------------------------------- */
/* liberation d'une GenomeSequence              */
/* -------------------------------------------- */

void FreeGenomeSequence(GenomeSeqPtr p_gen)
{
   if (p_gen) {

      if (p_gen->ope_kept)
	 free(p_gen->ope_kept);

      if (p_gen->fseqptr)
	 FreeFastaSequence(p_gen->fseqptr);

      free(p_gen);
   }

   return;
}
 
/* -------------------------------------------- */
/* lecture/redimensionnement d'un p_gen        */
/* <Lecture Serie>                              */
/* -------------------------------------------- */

Bool ReadGenomeSequence(FILE *streamin, GenomeSeqPtr p_gen, int standard_fasta_format) 
{

   if (p_gen == NULL) {
       MyError(stderr, 0, ERROR_TAG,
	 "ReadGenomeSequence:p_gen is NULL");
       return Faux;
   } 

   if (ReadFastaSequence(streamin, p_gen->fseqptr) == Faux) {
       return Faux;
   }

   if (GenomeSeqParseInfo(p_gen, standard_fasta_format) == Faux) {
     MyError(stderr, 0, ERROR_TAG,
       "ReadGenomeSequence:cannot parse %s", p_gen->fseqptr->comment);
     return Faux;
   }

   return Vrai;

}

/* -------------------------------------------- */
/* WriteGenomeSequence: ecrit la sequence	*/
/* de genome sur stream				*/
/* -------------------------------------------- */
void WriteGenomeSequence(FILE *stream, GenomeSeqPtr p_gen, int len_line)
{

   (void)fprintf(stream, ">%s %d %d %d %c\n", p_gen->mnemo, p_gen->rel_number,
		p_gen->from, p_gen->to, p_gen->strand);

   WriteFastaSequence(stream, p_gen->fseqptr, len_line, ONLY_SEQ);

   return;

}

/* -------------------------------------------- */
/* ComparScoreOpeKept: compares 		*/
/* scores of OpeKept ope 1 and ope2		*/
/* struct ptr for use in qsort			*/
/* return 0: if ope1.score == ope2.score	*/
/*       +1: if ope1.score <  ope2.score	*/
/*       -1: if ope1.score >  ope2.score	*/
/* Note: score is similarity score, the best is	*/
/* the largest BUT we want to sort in		*/
/* descending order				*/
/* -------------------------------------------- */
static int ComparScoreOpeKept(const void *ope1, const void *ope2)
{
   if ((int)(((OpeKeptPtr)ope1)->score) <
       (int)(((OpeKeptPtr)ope2)->score)   ) 

      return +1;

   else if ((int)(((OpeKeptPtr)ope1)->score) >
	    (int)(((OpeKeptPtr)ope2)->score)  )

     return -1;

   else

     return 0;

}

/* -------------------------------------------- */
/* Sort all kept ope of gene p_gen in		*/
/* decreasing score order			*/
/* -------------------------------------------- */
void sort_kept_ope_by_score_in_gene(GenomeSeqPtr p_gen)
{
	int	(*compar) (const void *, const void *);

	compar = ComparScoreOpeKept;

	qsort((void *)(p_gen->ope_kept), (size_t) (p_gen->actual_ope_kept),
			(size_t) sizeof(*(p_gen->ope_kept)), compar);
	return;
}
